
from typing import Any
from ..mesh3d import Mesh3D
from ..geometry import GeoObject
from ..selection import Selection
from ..physics.microwave.microwave_3d import Microwave3D
from ..settings import Settings

class DataSet:
    def __init__(self):
        pass

class SimState:
    
    def __init__(self, settings: Settings, params: dict[str, float]):
        self.params: dict[str, float] = params
        self.set: Settings = settings
        self.mesh: Mesh3D = Mesh3D()
        self.selections: dict[str, Selection] = dict()
        self.geos: dict[str, GeoObject] = dict()
        self.tasks: list[Any] = []
        self.data: DataSet = DataSet()
        self.mw: Microwave3D = Microwave3D()
        
class SimData:
    
    def __init__(self):
        self.states: list[SimState] = []


# EMerge is an open source Python based FEM EM simulation module.
# Copyright (C) 2025  Robert Fennis.

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, see
# <https://www.gnu.org/licenses/>.

from __future__ import annotations


from .geometry import GeoObject
from .geo.modeler import Modeler
from .physics.microwave.microwave_3d import Microwave3D
from .mesh3d import Mesh3D
from .mesher import Mesher
from .dataset import SimulationDataset
from .logsettings import LOG_CONTROLLER, DEBUG_COLLECTOR
from .plot.pyvista import PVDisplay
from .periodic import PeriodicCell
from .cacherun import get_build_section, get_run_section
from .settings import DEFAULT_SETTINGS, Settings
from .solver import EMSolver, Solver
from .simstate import SimState
from .selection import Selector, Selection
from typing import Literal, Generator, Any
from loguru import logger
import numpy as np
import gmsh # type: ignore
import os
import inspect
from pathlib import Path
import joblib
from atexit import register
import signal
from .. import __version__

############################################################
#                   EXCEPTION DEFINITIONS                  #
############################################################

_GMSH_ERROR_TEXT = """
--------------------------
Known problems/solutions:
(1) - PLC Error:  A segment and a facet intersect at point
    This can be caused when approximating thin curved volumes. Try to decrease the mesh size for that region.
--------------------------
"""

class SimulationError(Exception):
    pass

class VersionError(Exception):
    pass

############################################################
#                 BASE 3D SIMULATION MODEL                 #
############################################################


class Simulation:

    def __init__(self, 
                 modelname: str, 
                 loglevel: Literal['TRACE','DEBUG','INFO','WARNING','ERROR'] = 'INFO',
                 load_file: bool = False,
                 save_file: bool = False,
                 write_log: bool = False,
                 path_suffix: str = ".EMResults"):
        """Generate a Simulation class object.

        As a minimum a file name should be provided. Additionally you may provide it with any
        class that inherits from BaseDisplay. This will then be used for geometry displaying.

        Args:
            modelname (str): The name of the simulation model. This will be used for filenames and path names when saving data.
            loglevel ("DEBUG","INFO","WARNING","ERROR", optional): The loglevel to use for loguru. Defaults to 'INFO'.
            load_file (bool, optional): If the simulatio model should be loaded from a file. Defaults to False.
            save_file (bool, optional): if the simulation file should be stored to a file. Defaults to False.
            write_log (bool, optional): If a file should be created that contains the entire log of the simulation. Defaults to False.
            path_suffix (str, optional): The suffix that will be added to the results directory. Defaults to ".EMResults".
        """

        caller_file = Path(inspect.stack()[1].filename).resolve()
        base_path = caller_file.parent

        self.modelname = modelname
        self.modelpath = base_path / (modelname.lower()+path_suffix)
        self.mesher: Mesher = Mesher()
        self.modeler: Modeler = Modeler()
        
        self.state: SimState = SimState(self.modelname)
        self.select: Selector = Selector()
        self.settings: Settings = DEFAULT_SETTINGS

        ## Display
        self.display: PVDisplay = PVDisplay(self.state)
        
        ## STATES
        self.__active: bool = False
        self._defined_geometries: bool = False
        self._cell: PeriodicCell | None = None
        self.save_file: bool = save_file
        self.load_file: bool = load_file
        self._cache_run: bool = False
        self._file_lines: str = ''
        self._saved: bool = False
        
        ## Physics
        self.mw: Microwave3D = Microwave3D(self.state, self.mesher, self.settings)

        self._initialize_simulation()

        self.set_loglevel(loglevel)
        
        if write_log:
            self.set_write_log()

        LOG_CONTROLLER._flush_log_buffer()
        LOG_CONTROLLER._sys_info()

        self.__post_init__()

    ############################################################
    #                       PRIVATE FUNCTIONS                  #
    ############################################################

    @property
    def data(self) -> SimulationDataset:
        return self.state.data
    
    @property
    def mesh(self) -> Mesh3D:
        return self.state.mesh
    
    def __post_init__(self):
        pass

    def __setitem__(self, name: str, value: Any) -> None:
        """Store data in the current data container"""
        self.data.sim[name] = value

    def __getitem__(self, name: str) -> Any:
        """Get the data from the current data container"""
        return self.data.sim[name]
    
    def __enter__(self) -> Simulation:
        """This method is depricated with the new atexit system. It still exists for backwards compatibility.

        Returns:
            Simulation: the Simulation object
        """
        return self

    def __exit__(self, type, value, tb):
        """This method no longer does something. It only serves as backwards compatibility."""
        self._exit_gmsh()
        return False
    
    def _install_signal_handlers(self):
        # on SIGINT (Ctrl-C) or SIGTERM, call our exit routine
        for sig in (signal.SIGINT, signal.SIGTERM):
            signal.signal(sig, self._handle_signal)
    
    def _handle_signal(self, signum, frame):
        """
        Signal handler: do our cleanup, then re-raise
        the default handler so that exit code / traceback
        is as expected.
        """
        try:
            # run your atexit-style cleanup
            self._exit_gmsh()
        except Exception:
            # log but don’t block shutdown
            logger.exception("Error during signal cleanup")
        finally:
            # restore default handler and re‐send the signal
            signal.signal(signum, signal.SIG_DFL)
            os.kill(os.getpid(), signum)

    def _autosave(self):
        """Called by atexit as an emergency matter.
        """
        if not self.settings.auto_save:
            return
        
        if not self.settings.save_after_sim:
            if not self.mw._completed:
                return
            
        self._exit_gmsh()
        
    def _initialize_simulation(self):
        """Initializes the Simulation data and GMSH API with proper shutdown routines.
        """
        self.state.init()
        
        # If GMSH is not yet initialized (Two simulation in a file)
        if gmsh.isInitialized() == 0:
            logger.debug('Initializing GMSH')
            
            gmsh.initialize()
            # Set an exit handler for Ctrl+C cases
            self._install_signal_handlers()

            # Restier the Exit GMSH function on proper program abortion
            register(self._autosave)
        else:
            gmsh.finalize()
            gmsh.initialize()
            
        # Create a new GMSH model or load it
        if not self.load_file:
            gmsh.model.add(self.modelname)
        else:
            self.load()

        # Set the Simulation state to active
        self.__active = True
        return self

    def _exit_gmsh(self):
        # If the simulation object state is still active (GMSH is running)
        if not self.__active:
            return
        
        logger.debug('Exiting program')
        
        if DEBUG_COLLECTOR.any_warnings:
            logger.warning('EMerge simulation warnings:')
        for i, report in DEBUG_COLLECTOR.all_reports():
            logger.warning(f'{i}: {report}')
        
        # Save the file first
        if self.save_file:
            self.save(_force_save=False)
            
        # Finalize GMSH
        if gmsh.isInitialized():
            gmsh.finalize()

        logger.debug('GMSH Shut down successful')
        # set the state to active
        self.__active = False
        
    
    ############################################################
    #                       PUBLIC FUNCTIONS                  #
    ############################################################

    def cache_build(self) -> bool:
        """Checks if all the lines inside this if statement block are the same as those
        stored from a previous run. If so, then it returns false. Else it returns True.
        
        Can be used to capture an entire model simulation.
        
        Example:
        
        >>> if model.cache_build():
        >>>     box = em.geo.Box(...)
        >>>     # Other lines
        >>>     model.mw.run_sweep()
        >>> data = model.data.mw

        Returns:
            bool: If the code is not the same
        """
        
        self.save_file = True
        self._cache_run = True
        filestr = get_build_section()
        self._file_lines = filestr
        cachepath = self.modelpath / 'pylines.txt'
        
        # If there is not pylines file, simulate (can't have been run).
        if not cachepath.exists():
            logger.info('No cached data detected, running file')
            return True
        
        with open(cachepath, 'r') as file:
            lines = file.read()
        
        if lines==filestr:
            logger.info('Cached data detected! Loading data!')
            self.load()
            return False
        logger.info('Different cached data detected, rebuilding file.')
        return True
    
    def cache_run(self) -> bool:
        """Checks if all the lines before this call are the same as the lines
        stored from a previous run. If so, then it returns false. Else it returns True.
        
        Can be used to capture a run_sweep() call.
        
        Example:
        
        >>> if model.cache_run():
        >>>     model.mw.run_sweep()
        >>> data = model.data.mw

        Returns:
            bool: If the code is not the same
        """
        
        self._cache_run = True
        filestr = get_run_section()
        self._file_lines = filestr
        cachepath = self.modelpath / 'pylines.txt'
        
        # If there is not pylines file, simulate (can't have been run).
        if not cachepath.exists():
            self.save_file = True
            logger.info('No cached data detected, running simulation!')
            return True
        
        with open(cachepath, 'r') as file:
            lines = file.read()
        
        if lines==filestr:
            logger.info('Cached data detected! Loading data!')
            self.load()
            return False
        logger.info('Different cached data detected, rerunning simulation.')
        return True
    
    def check_version(self, target_version: str, *, log: bool = False) -> None:
        """
        Ensure the script targets an EMerge major.minor compatible with the current runtime.

        Only major.minor is enforced. Patch differences are ignored.

        Parameters
        ----------
        target_version : str
            The EMerge version this script was written for (e.g. "1.4.0" or "1.4").
        log : bool, optional
            If True and a `logger` is available, emit a single WARNING with the same
            message as the exception. Defaults to False.

        Raises
        ------
        VersionError
            If the script's target major.minor differs from the running EMerge major.minor.
        """
        try:
            from packaging.version import Version as _V
            def _mm(v: str):
                pv = _V(v)
                return (pv.major, pv.minor)
            script_mm = _mm(target_version)
            runtime_mm = _mm(__version__)
            newer = script_mm > runtime_mm
            older = script_mm < runtime_mm
        except Exception:
            def _parse_mm(v: str):
                parts = v.split(".")
                try:
                    major = int(parts[0])
                    minor = int(parts[1]) if len(parts) > 1 else 0
                    return (major, minor)
                except Exception:
                    # Fallback: compare as strings to avoid crashing the check itself
                    major = parts[0] if parts else "0"
                    minor = parts[1] if len(parts) > 1 else "0"
                    return (str(major), str(minor))
            script_mm = _parse_mm(target_version)
            runtime_mm = _parse_mm(__version__)
            newer = script_mm > runtime_mm
            older = script_mm < runtime_mm

        if not newer and not older:
            return  # major.minor match

        if newer:
            msg = (
                f"Script targets EMerge {target_version} (major.minor {script_mm[0]}.{script_mm[1]}), "
                f"but runtime is {__version__} (major.minor {runtime_mm[0]}.{runtime_mm[1]}). "
                "The script may rely on features added after your installed major.minor. "
                "Recommended: upgrade EMerge (`pip install --upgrade emerge`). "
                "If you know the script is compatible, you may remove this check."
            )
        else:  # older
            msg = (
                f"Script targets EMerge {target_version} (major.minor {script_mm[0]}.{script_mm[1]}), "
                f"but runtime is {__version__} (major.minor {runtime_mm[0]}.{runtime_mm[1]}). "
                "APIs may have changed since the targeted major.minor. "
                "Recommended: update the script for the current EMerge, or run a matching older release. "
                "If you know the script is compatible, you may remove this check."
            )

        if log:
            try:
                logger.warning(msg)
            except Exception:
                pass

        raise VersionError(msg)

    def activate(self, _indx: int | None = None, **variables) -> Simulation:
        """Searches for the permutaions of parameter sweep variables and sets the current geometry to the provided set."""
        self.state.activate(_indx, **variables)
        return self
    
    def save(self, _force_save: bool = True) -> None:
        """Saves the current model in the provided project directory."""
        # Ensure directory exists
        
        if self._saved and not _force_save:
            logger.debug('File already saved. Terminating save procedure.')
            return
        
        if not self.modelpath.exists():
            self.modelpath.mkdir(parents=True, exist_ok=True)
            logger.info(f"Created directory: {self.modelpath}")

        # Save mesh
        mesh_path = self.modelpath / 'mesh.msh'
        brep_path = self.modelpath / 'model.brep'

        gmsh.option.setNumber('Mesh.SaveParametric', 1)
        gmsh.option.setNumber('Mesh.SaveAll', 1)
        gmsh.model.geo.synchronize()
        gmsh.model.occ.synchronize()

        gmsh.write(str(mesh_path))
        gmsh.write(str(brep_path))
        logger.info(f"Saved mesh to: {mesh_path}")

        # Pack and save data
        dataset = self.state.get_dataset()
        data_path = self.modelpath / 'simdata.emerge'
        
        joblib.dump(dataset, str(data_path))
        
        if self._cache_run:
            cachepath = self.modelpath / 'pylines.txt'
            with open(str(cachepath), 'w') as f_out:
                f_out.write(self._file_lines)
            
        logger.info(f"Saved simulation data to: {data_path}")
        self._saved = True

    def load(self) -> None:
        """Loads the model from the project directory."""
        mesh_path = self.modelpath / 'mesh.msh'
        brep_path = self.modelpath / 'model.brep'
        data_path = self.modelpath / 'simdata.emerge'

        if not mesh_path.exists() or not data_path.exists():
            raise FileNotFoundError("Missing required mesh or data file.")

        # Load GMSH Mesh (Ideally Id remove)
        gmsh.open(str(brep_path))
        gmsh.merge(str(mesh_path))
        gmsh.model.geo.synchronize()
        gmsh.model.occ.synchronize()
        
        logger.info(f"Loaded mesh from: {mesh_path}")
        datapack = joblib.load(str(data_path))
        self.state.load_dataset(datapack)
        self.state.activate(0)
        
        logger.info(f"Loaded simulation data from: {data_path}")
    
    def set_loglevel(self, loglevel: Literal['DEBUG','INFO','WARNING','ERROR']) -> None:
        """Set the loglevel for loguru.

        Args:
            loglevel ('DEBUG','INFO','WARNING','ERROR'): The loglevel
        """
        logger.trace(f'Setting loglevel to {loglevel}')
        LOG_CONTROLLER.set_std_loglevel(loglevel)
        if loglevel not in ('TRACE'):
            gmsh.option.setNumber("General.Terminal", 0)

    def set_write_log(self) -> None:
        """Adds a file output for the logger."""
        logger.trace(f'Writing log to path = {self.modelpath}')
        LOG_CONTROLLER.set_write_file(self.modelpath)
        
    def view(self, 
             selections: list[Selection] | None = None, 
             use_gmsh: bool = False,
             plot_mesh: bool = False,
             volume_mesh: bool = True,
             opacity: float | None = None,
             labels: bool = False,
             bc: bool = False,
             bw: bool = False,
             face_labels: bool = False) -> None:
        """View the current geometry in either the BaseDisplay object (PVDisplay only) or
        the GMSH viewer.

        Args:
            selections (list[Selection] | None, optional): Optional selections to highlight. Defaults to None.
            use_gmsh (bool, optional): If GMSH's GUI should be used. Defaults to False.
            plot_mesh (bool, optional): If the mesh should be plot instead of the object. Defaults to False.
            volume_mesh (bool, optional): If the internal mesh should be plot instead of only the surface boundary mesh. Defaults to True
            opacity (float | None, optional): The object/mesh opacity. Defaults to None.
            labels: (bool, optional): If geometry name labels should be shown. Defaults to False.
            bc: (bool, optional): If you wish to show boundary condition selections in the view
            bw: (bool, optional): If you want to view in black-white mode.
        """
        if not (self.display is not None and self.mesh.defined) or use_gmsh:
            gmsh.model.occ.synchronize()
            gmsh.fltk.run()
            
            return
        if bw:
            self.display.drawing_bw()
            
        for geo in self.state.current_geo_state:
            self.display.add_object(geo, mesh=plot_mesh, opacity=opacity, volume_mesh=volume_mesh, label=labels)
            
            if face_labels and geo.dim==3:
                for face_name in geo._face_pointers.keys():
                    if geo.face(face_name).invalid:
                        continue
                    self.display.add_object(geo.face(face_name), color='yellow', opacity=0.1, label=face_name)
        if selections:
            [self.display.add_object(sel, color='red', opacity=0.6, label=sel.name) for sel in selections]
        if bc:
            for bc in self.mw.bc.boundary_conditions:
                if bc.selection.invalid:
                    continue
                self.display.add_object(bc.selection, color=bc._color, opacity=0.4, label=True, label_text=bc._name, texture=bc._texture)
        self.display.show()

        return None
        
    def set_periodic_cell(self, cell: PeriodicCell):
        """Set the given periodic cell object as the simulations peridicity.

        Args:
            cell (PeriodicCell): The PeriodicCell class
        """
        logger.trace(f'Setting {cell} as periodic cell object')
        self.mw.bc._cell = cell
        self._cell = cell

    def set_resolution(self, resolution: float) -> Simulation:
        """Sets the discretization resolution in the various physics interfaces.
        

        Args:
            resolution (float): The resolution as a float. Lower resolution is a finer mesh 

        Returns:
            Simulation: _description_
        """
        self.mw.set_resolution(resolution)
        
    def commit_geometry(self, *geometries: GeoObject | list[GeoObject]) -> None:
        """Finalizes and locks the current geometry state of the simulation.

        The geometries may be provided (legacy behavior) but are automatically managed in the background.
        
        """
        logger.trace('Committing final geometry.')
        self.state.store_geometry_data()
        
        logger.trace(f'Parsed geometries = {self.state.geos}')
        
        self.mesher.submit_objects(self.state.current_geo_state)
        
        self._defined_geometries = True
        self.display._facetags = [dt[1] for dt in gmsh.model.get_entities(2)]
    
    def all_geos(self) -> list[GeoObject]:
        """Returns all geometries in a list

        Returns:
            list[GeoObject]: A list of all GeoObjects
        """
        return self.state.current_geo_state
        
    def generate_mesh(self, regenerate: bool = False) -> None:
        """Generate the mesh. 
        This can only be done after commit_geometry(...) is called and if frequencies are defined.

        Args:
            name (str, optional): The mesh file name. Defaults to "meshname.msh".

        Raises:
            ValueError: ValueError if no frequencies are defined.
        """
        logger.info('Starting mesh generation phase.')
        if not regenerate:
            
            if not self._defined_geometries:
                self.commit_geometry()
            
            logger.trace(' (1) Installing periodic boundaries in mesher.')
            # Set the cell periodicity in GMSH
            if self._cell is not None:
                self.mesher.set_periodic_cell(self._cell)
            
            self.mw._initialize_bcs(self.state.manager.get_surfaces())

            # Check if frequencies are defined: TODO: Replace with a more generic check
            if self.mw.frequencies is None:
                raise ValueError('No frequencies defined for the simulation. Please set frequencies before generating the mesh.')

        gmsh.model.occ.synchronize()

        # Set the mesh size
        self.mesher._configure_bc_size(self.mw.bc.boundary_conditions)
        self.mesher._configure_mesh_size(self.mw.get_discretizer(), self.mw.resolution) # This makes no sense to do this here
        
        # Validity check
        x1, y1, z1, x2, y2, z2 = gmsh.model.getBoundingBox(-1, -1)
        bb_volume = (x2-x1)*(y2-y1)*(z2-z1)
        wl = 299792458/self.mw.frequencies[-1]
        Nelem = int(5 * bb_volume / (wl**3))
        if Nelem > 100_000 and DEFAULT_SETTINGS.size_check:
            DEBUG_COLLECTOR.add_report(f'An estimated {Nelem} tetrahedra are required for the bounding box of the geometry. This may imply a simulation domain that is very large.' + 
                                       'To disable this message. Set the .size_check parameter in model.settings to False.')
            
            raise SimulationError('Simulation requires too many elements.')
        logger.trace(' (2) Calling GMSH mesher')
        try:
            gmsh.logger.start()
            gmsh.model.mesh.generate(3)
            logs = gmsh.logger.get()
            gmsh.logger.stop()
            for log in logs:
                logger.trace('[GMSH] ' + log)
        except Exception:
            logger.error('GMSH Mesh error detected.')
            print(_GMSH_ERROR_TEXT)
            raise
        
        logger.info('GMSH Meshing complete!')
        self.mesh._pre_update(self.mesher._get_periodic_bcs())
        self.mesh.exterior_face_tags = self.mesher.domain_boundary_face_tags
        gmsh.model.occ.synchronize()
        self.state.store_geometry_data()
        logger.trace(' (3) Mesh routine complete')
        
    def parameter_sweep(self, clear_mesh: bool = True, **parameters: np.ndarray) -> Generator[tuple[float,...], None, None]:
        """Executes a parameteric sweep iteration.

        The first argument clear_mesh determines if the mesh should be cleared and rebuild in between sweeps. This is usually needed
        except when you change only port-properties or material properties. The parameters of the sweep can be provided as a set of 
        keyword arguments. As an example if you defined the axes: width=np.linspace(...) and height=np.linspace(...). You can
        add them as arguments using .parameter_sweep(True, width=width, height=height).

        The rest of the simulation commands should be inside the iteration scope

        Args:
            clear_mesh (bool, optional): Wether to clear the mesh in between sweeps. Defaults to True.

        Yields:
            Generator[tuple[float,...], None, None]: The parameters provided

        Example:
         >>> for W, H in model.parameter_sweep(True, width=widths, height=heights):
         >>>    // build simulation
         >>>    data = model.run_sweep()
         >>> // Extract the data
         >>> widths, heights, frequencies, S21 = data.ax('width','height','freq').S(2,1)
        """
        paramlist = sorted(list(parameters.keys()))
        dims = np.meshgrid(*[parameters[key] for key in paramlist], indexing='ij')
        dims_flat = [dim.flatten() for dim in dims]
        
        self.mw.cache_matrices = False
        logger.trace('Starting parameter sweep.')
        
        for i_iter in range(dims_flat[0].shape[0]):
            
            if clear_mesh and i_iter > 0:
                logger.info('Cleaning up mesh.')
                gmsh.clear()
                self.state.reset_geostate()
                self.mw.reset()
                
            
            params = {key: dim[i_iter] for key,dim in zip(paramlist, dims_flat)}
            
            self.state.set_parameters(params)
            
            logger.info(f'Iterating: {params}')
            if len(dims_flat)==1:
                yield dims_flat[0][i_iter]
            else:
                yield (dim[i_iter] for dim in dims_flat) # type: ignore
            
            
            if not clear_mesh:
                self.state.store_geometry_data()
        
        if not clear_mesh:
            self.state.store_geometry_data()
        
        self.mw.cache_matrices = True
        
        
    def export(self, filename: str):
        """Exports the model or mesh depending on the extension. 
        
        Exporting is realized by GMSH.
        Supported file formats are:
        
        3D Model: .opt, .geo_unrolled, .brep, .xao ,.step and .iges
        Mesh: .msh, .inp, .key, ._0000.rad, .celum, .cgns, .diff, .unv, .ir3, .mes, .mesh 
              .mail, .m, .bdf, .off, .p3d, .stl, .wrl, .vtk, .dat, .ply2, .su2, .neu, .x3d

        Args:
            filename (str): The filename
        """
        logger.trace(f'Writing geometry to {filename}')
        gmsh.write(filename)
        
    def set_solver(self, solver: EMSolver | Solver):
        """Set a given Solver class instance as the main solver. 
        Solvers will be checked on validity for the given problem.

        Args:
            solver (EMSolver | Solver): The solver objects
        """
        logger.trace(f'Setting solver to {solver}')
        self.mw.solveroutine.set_solver(solver)
        
    ############################################################
    #                     DEPRICATED FUNCTIONS                #
    ############################################################

    def define_geometry(self, *args):
        """DEPRICATED VERSION: Use .commit_geometry()
        """
        logger.warning('define_geometry() will be derpicated. Use commit_geometry() instead.')
        self.commit_geometry(*args)
    
    

    ############################################################
    #                   ADAPTIVE MESH REFINEMENT              #
    ############################################################
    def _reset_mesh(self):
        gmsh.model.mesh.clear()
        self.mw.reset(_reset_bc = False)
        self.state.reset_mesh()
    
    @staticmethod
    def guess_R(P: float, last_ratio: float = 1.0) -> float:
        # Coefficients for the refinement ratio calculation.
        
        a0 = 0.5
        c0 = 0.85
        x0 = 12
        q0 = (1-a0)*2/np.pi
        b0 = np.tan((c0-a0)/q0)/x0
        q0 = (0.8-a0)*2/np.pi
        
        ratio = (a0 + np.arctan(b0*P)*q0)
        if last_ratio > 1.0:
            ratio = ratio/0.9
        if last_ratio < 1.0:
            ratio = ratio*0.9
        
        return ratio
    
    @staticmethod
    def compute_ratio(new_point_percentage: float, ratios: np.ndarray, percentages: np.ndarray, P_target: float) -> float:
        """
        Strategy:
        - n=0: use guess_R(P_target, throttle=1.0)
        - n=1: use guess_R with throttle 0.5 if last P < target, else 2.0 if last P > 2*target,
                else keep the same R (already acceptable)
        - n=2: fit P = a1*(1/R) + a0 and solve for R; fallback to through-origin model if needed

        Returns R_guess in (0, 1].
        """
        #print(Rs, Ps)
        ratios = (np.asarray(ratios, dtype=float))
        percentages = (np.asarray(percentages, dtype=float))

        # Clean
        n = ratios.size
        
        if n == 1:
            last_R = float(ratios[-1])
            last_P = float(percentages[-1])
            
            if P_target <= last_P <= 2.0 * P_target:
                # Already acceptable
                return last_R * ((1.5*P_target)/last_P)**0.2
            
            if last_P > P_target*2.0:
                return last_R / 0.8
            else:
                return last_R * 0.8

        P_target = P_target*1.5
        x = 1.0 / ratios
        y = percentages
        dy = y - P_target

        # Indices by side of target
        idx_lo = np.where(dy < 0)[0]  # below target
        idx_hi = np.where(dy > 0)[0]  # above target
        idx_eq = np.where(dy == 0)[0]

        # Exact hit
        if idx_eq.size > 0:
            # Return corresponding R (already perfect)
            R_exact = float(ratios[idx_eq[0]])
            return float(np.clip(R_exact, 1e-6, 1.0))

        def pick_two(indices, reverse=False):
            # pick two closest to target among given indices
            order = np.argsort(np.abs(dy[indices]))
            if reverse:
                # when all below, we want the two with largest y (closest from below)
                order = np.argsort(-y[indices])
            return indices[order[:2]]

        # Choose two points
        if idx_lo.size > 0 and idx_hi.size > 0:
            # Bracket: closest below and closest above
            i_lo = idx_lo[np.argmin(np.abs(dy[idx_lo]))]
            i_hi = idx_hi[np.argmin(np.abs(dy[idx_hi]))]
            i1, i2 = i_lo, i_hi
        elif idx_lo.size == 0:
            # All above: take two smallest-above (closest to target)
            i12 = pick_two(idx_hi)  # by |dy|
            if i12.size < 2 and idx_hi.size >= 2:
                i12 = idx_hi[:2]
            i1, i2 = i12[0], i12[1]
        elif idx_hi.size == 0:
            # All below: take two largest-below (closest from below)
            i12 = pick_two(idx_lo, reverse=True)
            if i12.size < 2 and idx_lo.size >= 2:
                i12 = idx_lo[:2]
            i1, i2 = i12[0], i12[1]
        else:
            # Fallback (should not trigger)
            i1, i2 = 0, 1

        x1, y1 = float(x[i1]), float(y[i1])
        x2, y2 = float(x[i2]), float(y[i2])

        # Linear solve in y(x): y = y1 + (y2-y1)/(x2-x1) * (x - x1)
        # Target xt: xt = x1 + (P_target - y1) * (x2 - x1) / (y2 - y1)
        if abs(y2 - y1) > 1e-16:
            xt = x1 + (P_target - y1) * (x2 - x1) / (y2 - y1)
        else:
            # Degenerate: same y; choose midpoint in x
            xt = 0.5 * (x1 + x2)

        # Ensure positive xt
        xt = float(max(xt, 1e-16))
        R_new = 1.0 / xt

        # Clamp to (0, 1]
        return float(np.clip(R_new, 1e-6, 1.0))

    def adaptive_mesh_refinement(self, 
                                 max_steps: int = 6,
                                 min_refined_passes: int = 1,
                                 convergence: float = 0.02,
                                 magnitude_convergence: float = 2.0,
                                 phase_convergence: float = 180,
                                 max_tets: int = 1e6,
                                 refinement_ratio: float = 0.6,
                                 growth_rate: float = 1.6,
                                 minimum_refinement_percentage: float = 20.0, 
                                 error_field_inclusion_percentage: float = 50.0,
                                 minimum_steps: int = 1,
                                 frequency: float | list[float] = None,
                                 show_mesh: bool = False) -> SimulationDataset:
        """ A beta-version of adaptive mesh refinement.

        Convergence Criteria:
            (1): max(abs(S[n]-S[n-1]))
            (2): max(abs(abs(S[n]) - abs(S[n-1])))
            (3): max(angle(S[n]/S[n-1])) * 180/π
        
        Args:
            max_steps (int, optional): The maximum number of refinement steps. Defaults to 6.
            min_refined_passes (int, optional): The minimum number of refined passes. Defaults to 1.
            convergence (float, optional): The S-paramerter convergence (1). Defaults to 0.02.
            magnitude_convergence (float, optional): The S-parameter magnitude convergence (2). Defaults to 2.0.
            phase_convergence (float, optional): The S-parameter Phase convergence (3). Defaults to 180.
            refinement_ratio (float, optional): The size reduction of mesh elements by original length. Defaults to 0.75.
            growth_rate (float, optional): The mesh size growth rate. Defaults to 3.0.
            minimum_refinement_percentage (float, optional): The minimum mesh size increase . Defaults to 15.0.
            error_field_inclusion_percentage (float, optional): A percentage of tet elements to be included for refinement. Defaults to 5.0.
            minimum_steps (int, optional): The minimum number of adaptive steps to execute. Defaults to 1.
            frequency (float, optional): The refinement frequency. Defaults to None.
            show_mesh (bool, optional): If the intermediate meshes should be shown (freezes simulation). Defaults to False

        Returns:
            SimulationDataset: _description_
        """
        from .physics.microwave.adaptive_mesh import select_refinement_indices, reduce_point_set, compute_convergence, tet_to_node
        
        max_freq = np.max(self.mw.frequencies)
        
        if frequency is not None:
            sim_freqs = frequency
            if isinstance(sim_freqs, float):
                sim_freqs = [sim_freqs,]
        else:
            sim_freqs = [max_freq,]
        
        
        S_matrices: list[list[np.ndarray]] = []

        last_n_tets: int = self.mesh.n_tets
        logger.info(f'Initial mesh has {last_n_tets} tetrahedra')  
        
        passed = 0
        
        self.state.stash()
        
        NF = len(sim_freqs)
        
        original_ratio = refinement_ratio
        for step in range(1,max_steps+1):
            
            self.data.sim.new(iter_step=step)
            
            datas = []
            fields = []
            Smats = []
            
            for sf in sim_freqs:
                data, solve_ids = self.mw._run_adaptive_mesh(step, sf)
                datas.append(data)
                fields.append(data.field[-1])
                Smats.append(data.scalar[-1].Sp)
                
            S_matrices.append(Smats)
            
            if step > minimum_steps:
                S0s = S_matrices[-2]
                S1s = S_matrices[-1]
                
                max_complx = 0
                max_mag = 0
                max_phase = 0
                
                for i in range(NF):
                    conv_complex, conv_mag, conv_phase = compute_convergence(S0s[i], S1s[i])
                    max_complx = max(max_complx, conv_complex)
                    max_mag = max(max_mag, conv_mag)    
                    max_phase = max(max_phase, conv_phase)
                    
                logger.info(f'Pass {step}: Convergence = {max_complx:.3f}, Mag = {max_mag:.3f}, Phase = {max_phase:.1f} deg')
                
                if max_complx <= convergence and max_phase < phase_convergence and max_mag < magnitude_convergence:
                    logger.info(f'Pass {step}: Mesh refinement passed!')
                    passed += 1
                else:
                    passed = 0
            
            if passed >= min_refined_passes and step > minimum_steps:
                logger.info(f'Adaptive mesh refinement successfull with {self.mesh.n_tets} tetrahedra.')
                break
            
            errors = np.empty((self.mesh.n_tets, NF), dtype=np.float64)
            for i in range(NF):
                error, lengths = fields[i]._solution_quality(solve_ids)
                errors[:,i] = error
            
            
            error = np.max(errors, axis=1)
                
            idx = select_refinement_indices(error, error_field_inclusion_percentage/100)
            idx = idx[::-1]
            
            npts = idx.shape[0]
            np_percentage = npts/self.mesh.n_tets * 100
            
            original_ratio = 0.75*original_ratio + 0.25*refinement_ratio
            refinement_ratio = original_ratio
            Ratios = []
            Percentages = []
            
            
            included = np.zeros((self.mesh.n_tets, ), dtype=np.bool_)
            included[idx] = True
            
            coords, sizes = tet_to_node(self.mesh.nodes, self.mesh.tets, lengths, included)
            self.mesher.add_refinement_points(coords, sizes, refinement_ratio*np.ones_like(sizes))#self.mw.mesh.centers[:,idx], lengths[idx], refinement_ratio*np.ones_like(lengths[idx]))
                
            new_ids = reduce_point_set(self.mesher._amr_coords, growth_rate, self.mesher._amr_sizes, refinement_ratio, 0.20)
            
            nremoved = self.mesher._amr_coords.shape[1] - len(new_ids)
            
            logger.info(f'    Pass {step}: Added {len(sizes) - nremoved} new refinement points with ratio {refinement_ratio}.')
            
            self.mesher._amr_coords = self.mesher._amr_coords[:,new_ids]
            self.mesher._amr_sizes = self.mesher._amr_sizes[new_ids]
            self.mesher._amr_ratios = self.mesher._amr_ratios[new_ids]
            self.mesher._amr_new = self.mesher._amr_new[new_ids]
            
            logger.debug(f'    Initial refinement ratio: {refinement_ratio}')
            
            # Mesh refinement loop. Only escapes if the mesh refined a certain set percentage.
            counter = 0
            while True:
                counter += 1
                if counter == 10:
                    logger.warning('    More than 10 attempts at reaching the target refinement. Continuing with current.')
                    break
                counter += 1
                self._reset_mesh()
                self.mesher.set_refinement_function(growth_rate, 2.0)
                self.generate_mesh(True)
                percentage = (self.mesh.n_tets/last_n_tets - 1) * 100
                logger.info(f'    Pass {step}: New mesh has {self.mesh.n_tets} (+{percentage:.1f}%) tetrahedra.')  
                
                Ratios.append(refinement_ratio)
                Percentages.append(percentage)
                
                if len(Percentages) >= 2:
                    if abs(Percentages[-2]-Percentages[-1]) == 0.0:
                        logger.warning(f'No refinement realized, decreasing refinment ratio.')
                        refinement_ratio = refinement_ratio * 0.5
                        self.mesher.set_ratio(refinement_ratio)
                        continue
                
                if percentage < minimum_refinement_percentage or percentage > (minimum_refinement_percentage*2):
                    
                    refinement_ratio = self.compute_ratio(np_percentage, Ratios, Percentages, minimum_refinement_percentage)
                    logger.info(f'    Refinement target not reached! New ratio = {refinement_ratio:.3f}')
                    self.mesher.set_ratio(refinement_ratio)
                    if refinement_ratio >= 1.0:
                        logger.warning(f'Refinement ratio pushed above 1.0... continuing with current percentage.')
                        break
                    continue
                
                
                break
            
            last_n_tets = self.mesh.n_tets
            if show_mesh:
                self.view(plot_mesh=True, volume_mesh=True)
            
            if last_n_tets > max_tets:
                logger.warning(f'Aborting refinement because the number of tets exceeds the maximum: {last_n_tets}>{max_tets}')
                break
        if passed < min_refined_passes:
            logger.warning('Adaptive mesh refinement did not converge!')
            
        if show_mesh:
                self.view(plot_mesh=True, volume_mesh=True)
        old = self.state.reload()
        self.state.store_geometry_data()
        
        return old
    
class SimulationBeta(Simulation):
    
    
    def __post_init__(self):
        pass
        #self.mesher.set_algorithm(Algorithm3D.HXT)
        #logger.debug('Setting mesh algorithm to HXT')
        
    
    def __enter__(self) -> SimulationBeta:
        """This method is depricated with the new atexit system. It still exists for backwards compatibility.

        Returns:
            SimulationBeta: the SimulationBeta object
        """
        return self
    
    def _reset_mesh(self):
        gmsh.model.mesh.clear()
        self.mw.reset(_reset_bc = False)
        self.state.reset_mesh()
    
    @staticmethod
    def guess_R(P: float, last_ratio: float = 1.0) -> float:
        # Coefficients for the refinement ratio calculation.
        
        a0 = 0.5
        c0 = 0.85
        x0 = 12
        q0 = (1-a0)*2/np.pi
        b0 = np.tan((c0-a0)/q0)/x0
        q0 = (0.8-a0)*2/np.pi
        
        ratio = (a0 + np.arctan(b0*P)*q0)
        if last_ratio > 1.0:
            ratio = ratio/0.9
        if last_ratio < 1.0:
            ratio = ratio*0.9
        
        return ratio
    
    @staticmethod
    def compute_ratio(new_point_percentage: float, ratios: np.ndarray, percentages: np.ndarray, P_target: float) -> float:
        """
        Strategy:
        - n=0: use guess_R(P_target, throttle=1.0)
        - n=1: use guess_R with throttle 0.5 if last P < target, else 2.0 if last P > 2*target,
                else keep the same R (already acceptable)
        - n=2: fit P = a1*(1/R) + a0 and solve for R; fallback to through-origin model if needed

        Returns R_guess in (0, 1].
        """
        #print(Rs, Ps)
        ratios = (np.asarray(ratios, dtype=float))
        percentages = (np.asarray(percentages, dtype=float))

        # Clean
        n = ratios.size
        
        if n == 1:
            last_R = float(ratios[-1])
            last_P = float(percentages[-1])
            
            if P_target <= last_P <= 2.0 * P_target:
                # Already acceptable
                return last_R * ((1.5*P_target)/last_P)**0.2
            
            if last_P > P_target*2.0:
                return last_R / 0.8
            else:
                return last_R * 0.8

        P_target = P_target*1.5
        x = 1.0 / ratios
        y = percentages
        dy = y - P_target

        # Indices by side of target
        idx_lo = np.where(dy < 0)[0]  # below target
        idx_hi = np.where(dy > 0)[0]  # above target
        idx_eq = np.where(dy == 0)[0]

        # Exact hit
        if idx_eq.size > 0:
            # Return corresponding R (already perfect)
            R_exact = float(ratios[idx_eq[0]])
            return float(np.clip(R_exact, 1e-6, 1.0))

        def pick_two(indices, reverse=False):
            # pick two closest to target among given indices
            order = np.argsort(np.abs(dy[indices]))
            if reverse:
                # when all below, we want the two with largest y (closest from below)
                order = np.argsort(-y[indices])
            return indices[order[:2]]

        # Choose two points
        if idx_lo.size > 0 and idx_hi.size > 0:
            # Bracket: closest below and closest above
            i_lo = idx_lo[np.argmin(np.abs(dy[idx_lo]))]
            i_hi = idx_hi[np.argmin(np.abs(dy[idx_hi]))]
            i1, i2 = i_lo, i_hi
        elif idx_lo.size == 0:
            # All above: take two smallest-above (closest to target)
            i12 = pick_two(idx_hi)  # by |dy|
            if i12.size < 2 and idx_hi.size >= 2:
                i12 = idx_hi[:2]
            i1, i2 = i12[0], i12[1]
        elif idx_hi.size == 0:
            # All below: take two largest-below (closest from below)
            i12 = pick_two(idx_lo, reverse=True)
            if i12.size < 2 and idx_lo.size >= 2:
                i12 = idx_lo[:2]
            i1, i2 = i12[0], i12[1]
        else:
            # Fallback (should not trigger)
            i1, i2 = 0, 1

        x1, y1 = float(x[i1]), float(y[i1])
        x2, y2 = float(x[i2]), float(y[i2])

        # Linear solve in y(x): y = y1 + (y2-y1)/(x2-x1) * (x - x1)
        # Target xt: xt = x1 + (P_target - y1) * (x2 - x1) / (y2 - y1)
        if abs(y2 - y1) > 1e-16:
            xt = x1 + (P_target - y1) * (x2 - x1) / (y2 - y1)
        else:
            # Degenerate: same y; choose midpoint in x
            xt = 0.5 * (x1 + x2)

        # Ensure positive xt
        xt = float(max(xt, 1e-16))
        R_new = 1.0 / xt

        # Clamp to (0, 1]
        return float(np.clip(R_new, 1e-6, 1.0))

    def adaptive_mesh_refinement(self, 
                                 max_steps: int = 6,
                                 min_refined_passes: int = 1,
                                 convergence: float = 0.02,
                                 magnitude_convergence: float = 2.0,
                                 phase_convergence: float = 180,
                                 max_tets: int = 1e6,
                                 refinement_ratio: float = 0.6,
                                 growth_rate: float = 1.6,
                                 minimum_refinement_percentage: float = 20.0, 
                                 error_field_inclusion_percentage: float = 50.0,
                                 minimum_steps: int = 1,
                                 frequency: float | list[float] = None,
                                 show_mesh: bool = False) -> SimulationDataset:
        """ A beta-version of adaptive mesh refinement.

        Convergence Criteria:
            (1): max(abs(S[n]-S[n-1]))
            (2): max(abs(abs(S[n]) - abs(S[n-1])))
            (3): max(angle(S[n]/S[n-1])) * 180/π
        
        Args:
            max_steps (int, optional): The maximum number of refinement steps. Defaults to 6.
            min_refined_passes (int, optional): The minimum number of refined passes. Defaults to 1.
            convergence (float, optional): The S-paramerter convergence (1). Defaults to 0.02.
            magnitude_convergence (float, optional): The S-parameter magnitude convergence (2). Defaults to 2.0.
            phase_convergence (float, optional): The S-parameter Phase convergence (3). Defaults to 180.
            refinement_ratio (float, optional): The size reduction of mesh elements by original length. Defaults to 0.75.
            growth_rate (float, optional): The mesh size growth rate. Defaults to 3.0.
            minimum_refinement_percentage (float, optional): The minimum mesh size increase . Defaults to 15.0.
            error_field_inclusion_percentage (float, optional): A percentage of tet elements to be included for refinement. Defaults to 5.0.
            minimum_steps (int, optional): The minimum number of adaptive steps to execute. Defaults to 1.
            frequency (float, optional): The refinement frequency. Defaults to None.
            show_mesh (bool, optional): If the intermediate meshes should be shown (freezes simulation). Defaults to False

        Returns:
            SimulationDataset: _description_
        """
        from .physics.microwave.adaptive_mesh import select_refinement_indices, reduce_point_set, compute_convergence, tet_to_node
        
        max_freq = np.max(self.mw.frequencies)
        
        if frequency is not None:
            sim_freqs = frequency
            if isinstance(sim_freqs, float):
                sim_freqs = [sim_freqs,]
        else:
            sim_freqs = [max_freq,]
        
        
        S_matrices: list[list[np.ndarray]] = []

        last_n_tets: int = self.mesh.n_tets
        logger.info(f'Initial mesh has {last_n_tets} tetrahedra')  
        
        passed = 0
        
        self.state.stash()
        
        NF = len(sim_freqs)
        
        original_ratio = refinement_ratio
        for step in range(1,max_steps+1):
            
            self.data.sim.new(iter_step=step)
            
            datas = []
            fields = []
            Smats = []
            
            for sf in sim_freqs:
                data, solve_ids = self.mw._run_adaptive_mesh(step, sf)
                datas.append(data)
                fields.append(data.field[-1])
                Smats.append(data.scalar[-1].Sp)
                
            S_matrices.append(Smats)
            
            if step > minimum_steps:
                S0s = S_matrices[-2]
                S1s = S_matrices[-1]
                
                max_complx = 0
                max_mag = 0
                max_phase = 0
                
                for i in range(NF):
                    conv_complex, conv_mag, conv_phase = compute_convergence(S0s[i], S1s[i])
                    max_complx = max(max_complx, conv_complex)
                    max_mag = max(max_mag, conv_mag)    
                    max_phase = max(max_phase, conv_phase)
                    
                logger.info(f'Pass {step}: Convergence = {max_complx:.3f}, Mag = {max_mag:.3f}, Phase = {max_phase:.1f} deg')
                
                if max_complx <= convergence and max_phase < phase_convergence and max_mag < magnitude_convergence:
                    logger.info(f'Pass {step}: Mesh refinement passed!')
                    passed += 1
                else:
                    passed = 0
            
            if passed >= min_refined_passes and step > minimum_steps:
                logger.info(f'Adaptive mesh refinement successfull with {self.mesh.n_tets} tetrahedra.')
                break
            
            errors = np.empty((self.mesh.n_tets, NF), dtype=np.float64)
            for i in range(NF):
                error, lengths = fields[i]._solution_quality(solve_ids)
                errors[:,i] = error
            
            
            error = np.max(errors, axis=1)
                
            idx = select_refinement_indices(error, error_field_inclusion_percentage/100)
            idx = idx[::-1]
            
            npts = idx.shape[0]
            np_percentage = npts/self.mesh.n_tets * 100
            
            original_ratio = 0.75*original_ratio + 0.25*refinement_ratio
            refinement_ratio = original_ratio
            Ratios = []
            Percentages = []
            
            
            included = np.zeros((self.mesh.n_tets, ), dtype=np.bool_)
            included[idx] = True
            
            coords, sizes = tet_to_node(self.mesh.nodes, self.mesh.tets, lengths, included)
            self.mesher.add_refinement_points(coords, sizes, refinement_ratio*np.ones_like(sizes))#self.mw.mesh.centers[:,idx], lengths[idx], refinement_ratio*np.ones_like(lengths[idx]))
                
            new_ids = reduce_point_set(self.mesher._amr_coords, growth_rate, self.mesher._amr_sizes, refinement_ratio, 0.20)
            
            nremoved = self.mesher._amr_coords.shape[1] - len(new_ids)
            
            logger.info(f'    Pass {step}: Added {len(sizes) - nremoved} new refinement points with ratio {refinement_ratio}.')
            
            self.mesher._amr_coords = self.mesher._amr_coords[:,new_ids]
            self.mesher._amr_sizes = self.mesher._amr_sizes[new_ids]
            self.mesher._amr_ratios = self.mesher._amr_ratios[new_ids]
            self.mesher._amr_new = self.mesher._amr_new[new_ids]
            
            logger.debug(f'    Initial refinement ratio: {refinement_ratio}')
            
            # Mesh refinement loop. Only escapes if the mesh refined a certain set percentage.
            counter = 0
            while True:
                counter += 1
                if counter == 10:
                    logger.warning('    More than 10 attempts at reaching the target refinement. Continuing with current.')
                    break
                counter += 1
                self._reset_mesh()
                self.mesher.set_refinement_function(growth_rate, 2.0)
                self.generate_mesh(True)
                percentage = (self.mesh.n_tets/last_n_tets - 1) * 100
                logger.info(f'    Pass {step}: New mesh has {self.mesh.n_tets} (+{percentage:.1f}%) tetrahedra.')  
                
                Ratios.append(refinement_ratio)
                Percentages.append(percentage)
                
                if len(Percentages) >= 2:
                    if abs(Percentages[-2]-Percentages[-1]) == 0.0:
                        logger.warning(f'No refinement realized, decreasing refinment ratio.')
                        refinement_ratio = refinement_ratio * 0.5
                        self.mesher.set_ratio(refinement_ratio)
                        continue
                
                if percentage < minimum_refinement_percentage or percentage > (minimum_refinement_percentage*2):
                    
                    refinement_ratio = self.compute_ratio(np_percentage, Ratios, Percentages, minimum_refinement_percentage)
                    logger.info(f'    Refinement target not reached! New ratio = {refinement_ratio:.3f}')
                    self.mesher.set_ratio(refinement_ratio)
                    if refinement_ratio >= 1.0:
                        logger.warning(f'Refinement ratio pushed above 1.0... continuing with current percentage.')
                        break
                    continue
                
                
                break
            
            last_n_tets = self.mesh.n_tets
            if show_mesh:
                self.view(plot_mesh=True, volume_mesh=True)
            
            if last_n_tets > max_tets:
                logger.warning(f'Aborting refinement because the number of tets exceeds the maximum: {last_n_tets}>{max_tets}')
                break
        if passed < min_refined_passes:
            logger.warning('Adaptive mesh refinement did not converge!')
            
        if show_mesh:
                self.view(plot_mesh=True, volume_mesh=True)
        old = self.state.reload()
        self.state.store_geometry_data()
        
        return old
    
    


# EMerge is an open source Python based FEM EM simulation module.
# Copyright (C) 2025  Robert Fennis.

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, see
# <https://www.gnu.org/licenses/>.

from __future__ import annotations
from ...mesh3d import Mesh3D
from ...simstate import SimState
from ...geometry import GeoObject
from ...selection import FaceSelection, DomainSelection, EdgeSelection, Selection, encode_data
from ...physics.microwave.microwave_bc import PortBC, ModalPort

from emsutil.pyvista import EMergeDisplay, setdefault, cmap_names, _AnimObject
import numpy as np
from typing import Iterable, Literal

import pyvista as pv

def _min_distance(xs, ys, zs):
    """
    Compute the minimum Euclidean distance between any two points
    defined by the 1D arrays xs, ys, zs.
    
    Parameters:
        xs (np.ndarray): x-coordinates of the points
        ys (np.ndarray): y-coordinates of the points
        zs (np.ndarray): z-coordinates of the points
    
    Returns:
        float: The minimum Euclidean distance between any two points
    """
    # Stack the coordinates into a (N, 3) array
    points = np.stack((xs, ys, zs), axis=-1)

    # Compute pairwise squared distances using broadcasting
    diff = points[:, np.newaxis, :] - points[np.newaxis, :, :]
    dists_squared = np.sum(diff ** 2, axis=-1)

    # Set diagonal to infinity to ignore zero distances to self
    np.fill_diagonal(dists_squared, np.inf)

    # Get the minimum distance
    min_dist = np.sqrt(np.min(dists_squared))
    return min_dist

def _select(obj: GeoObject | Selection) -> Selection:
    if isinstance(obj, GeoObject):
        return obj.selection
    return obj

def _merge(lst: Iterable[GeoObject | Selection]) -> Selection:
    selections = [_select(item) for item in lst]
    dim = selections[0].dim
    all_tags = []
    for item in lst:
        all_tags.extend(_select(item).tags)
    
    if dim==1:
        return EdgeSelection(all_tags)
    elif dim==2:
        return FaceSelection(all_tags)
    elif dim==3:
        return DomainSelection(all_tags)
    else:
        return Selection(all_tags)

class PVDisplay(EMergeDisplay):

    def __post_init__(self, state: SimState):
        self._state: SimState = state
        self._selector._set_encoder_function(encode_data)
        
    def _get_edge_length(self):
        return max(1e-3, min(self._mesh.edge_lengths))
    ############################################################
    #                       SPECIFIC METHODS                  #
    ############################################################
    
    def _volume_edges(self, obj: GeoObject | Selection) -> pv.UnstructuredGrid:
        """Adds the edges of objects

        Args:
            obj (DomainSelection | None, optional): _description_. Defaults to None.

        Returns:
            pv.UnstructuredGrid: The unstrutured grid object
        """
        edge_ids = self._mesh.domain_edges(obj.dimtags)
        nedges = edge_ids.shape[0]
        cells = np.zeros((nedges,3), dtype=np.int64)
        cells[:,1:] = self._mesh.edges[:,edge_ids].T
        cells[:,0] = 2
        celltypes = np.full(nedges, fill_value=pv.CellType.CUBIC_LINE, dtype=np.uint8)
        points = self._mesh.nodes.copy().T
        return pv.UnstructuredGrid(cells, celltypes, points)
    
    def mesh_surface(self, surface: FaceSelection) -> pv.UnstructuredGrid:
        tris = self._mesh.get_triangles(surface.tags)
        ntris = tris.shape[0]
        cells = np.zeros((ntris,4), dtype=np.int64)
        cells[:,1:] = self._mesh.tris[:,tris].T
        cells[:,0] = 3
        celltypes = np.full(ntris, fill_value=pv.CellType.TRIANGLE, dtype=np.uint8)
        points = self._mesh.nodes.copy().T
        points[:,2] += self.set.z_boost
        return pv.UnstructuredGrid(cells, celltypes, points)
    
    def mesh(self, obj: GeoObject | Selection | Iterable) -> pv.UnstructuredGrid | None:
        if isinstance(obj, Iterable):
            obj = _merge(obj)
        else:
            obj = _select(obj)
        
        if isinstance(obj, DomainSelection):
            return self.mesh_volume(obj)
        elif isinstance(obj, FaceSelection):
            return self.mesh_surface(obj)
        else:
            return None


    ############################################################
    #                        EMERGE METHODS                    #
    ############################################################

    def mesh_volume(self, volume: DomainSelection) -> pv.UnstructuredGrid:
        tets = self._mesh.get_tetrahedra(volume.tags)
        ntets = tets.shape[0]
        cells = np.zeros((ntets,5), dtype=np.int64)
        cells[:,1:] = self._mesh.tets[:,tets].T
        cells[:,0] = 4
        celltypes = np.full(ntets, fill_value=pv.CellType.TETRA, dtype=np.uint8)
        points = self._mesh.nodes.copy().T
        return pv.UnstructuredGrid(cells, celltypes, points)
    
    
    @property
    def _mesh(self) -> Mesh3D:
        return self._state.mesh
    
    def add_object(self, obj: GeoObject | Selection, 
                   mesh: bool = False, 
                   volume_mesh: bool = True, 
                   label: bool = False, 
                   label_text: str | None = None, 
                   texture: str | None = None, *args, **kwargs):
        
        if isinstance(obj, GeoObject):
            if obj._hidden:
                return
        
        
        self._add_obj(self.mesh(obj), obj.dim, 
                      plot_mesh=mesh, 
                      volume_mesh=volume_mesh, 
                      metal=obj._metal, 
                      opacity=obj.opacity,
                      color=obj.color_rgb, 
                      texture=texture)

        self._plot.add_mesh(self._volume_edges(_select(obj)), line_width=self.set.theme.geo_edge_width, color=self.set.theme.geo_edge_color, show_edges=True)

        if label:
            points = []
            labels = []
            label_text = obj.name if label_text is None else label_text
            for dim, tag in obj.dimtags:
                if dim==2:
                    points.append(self._mesh.ftag_to_point[tag])
                else:
                    points.append(self._mesh.dimtag_to_center[(dim, tag)])
                labels.append(label_text)
            self._plot.add_point_labels(points, labels, shape_color=self.set.theme.label_color)
            
    def add_objects(self, *objects, **kwargs) -> None:
        """Add a series of objects provided as a list of arguments
        """
        for obj in objects:
            self.add_object(obj, **kwargs)
        
    def add_scatter(self, xs: np.ndarray, ys: np.ndarray, zs: np.ndarray):
        """Adds a scatter point cloud

        Args:
            xs (np.ndarray): The X-coordinate
            ys (np.ndarray): The Y-coordinate
            zs (np.ndarray): The Z-coordinate
        """
        cloud = pv.PolyData(np.array([xs,ys,zs]).T)
        self._data_sets.append(cloud)
        self._plot.add_points(cloud)

    def add_portmode(self, port: PortBC, 
                     Npoints: int = 10, 
                     dv=(0,0,0), 
                     XYZ=None,
                     field: Literal['E','H'] = 'E', 
                     k0: float | None = None,
                     mode_number: int | None = None) -> None:
        
        if XYZ:
            X,Y,Z = XYZ
        else:
            tris = self._mesh.get_triangles(port.selection.tags)
            ids = np.sort(np.unique(self._mesh.tris[:,tris].flatten()))
            X = self._mesh.tri_centers[0,tris]
            Y = self._mesh.tri_centers[1,tris]
            Z = self._mesh.tri_centers[2,tris]
            X2 = self._mesh.nodes[0,ids]
            Y2 = self._mesh.nodes[1,ids]
            Z2 = self._mesh.nodes[2,ids]
            X = np.concatenate((X,X2))
            Y = np.concatenate((Y,Y2))
            Z = np.concatenate((Z,Z2))
        
        X = X+dv[0]
        Y = Y+dv[1]
        Z = Z+dv[2]
        xf = X.flatten()
        yf = Y.flatten()
        zf = Z.flatten()

        d = _min_distance(xf, yf, zf)

        if port.vintline is not None:
            for line in port.vintline:
                xs, ys, zs = line.cpoint
                p_line = pv.Line(
                    pointa=(xs[0], ys[0], zs[0]),
                    pointb=(xs[-1], ys[-1], zs[-1]),
                )
                self._plot.add_mesh(
                    p_line,
                    color='red',
                    pickable=False,
                    line_width=3.0,
                )
            
        if k0 is None:
            if isinstance(port, ModalPort):
                k0 = port.get_mode(0).k0
            else:
                k0 = 1
        
        if isinstance(mode_number, int):
            port.selected_mode = mode_number
        
        F = port.port_mode_3d_global(xf,yf,zf,k0, which=field)

        Fx = F[0,:].reshape(X.shape).T
        Fy = F[1,:].reshape(X.shape).T
        Fz = F[2,:].reshape(X.shape).T

        if field=='H':
            F = np.imag(F.T)
            Fnorm = np.sqrt(Fx.imag**2 + Fy.imag**2 + Fz.imag**2).T
        else:
            F = np.real(F.T)
            Fnorm = np.sqrt(Fx.real**2 + Fy.real**2 + Fz.real**2).T
        
        if XYZ is not None:
            grid = pv.StructuredGrid(X,Y,Z)
            self.add_surf(X,Y,Z,Fnorm, _fieldname = 'portfield')
            self._wrap_plot(grid, scalars = Fnorm.T, opacity=0.8, pickable=False)

        Emag = F/np.max(Fnorm.flatten())*d*3
        actor = self._plot.add_arrows(np.array([xf,yf,zf]).T, Emag)
        self._data_sets.append(actor.mapper.dataset)
        
    def add_surf(self, 
                 x: np.ndarray,
                 y: np.ndarray,
                 z: np.ndarray,
                 field: np.ndarray,
                 scale: Literal['lin','log','symlog'] = 'lin',
                 cmap: cmap_names | None = None,
                 clim: tuple[float, float] | None = None,
                 opacity: float = 1.0,
                 symmetrize: bool = False,
                 _fieldname: str | None = None,
                 **kwargs,) -> pv.DataSet:
        """Add a surface plot to the display
        The X,Y,Z coordinates must be a 2D grid of data points. The field must be a real field with the same size.

        Args:
            x (np.ndarray): The X-grid array
            y (np.ndarray): The Y-grid array
            z (np.ndarray): The Z-grid array
            field (np.ndarray): The scalar field to display
            scale (Literal["lin","log","symlog"], optional): The colormap scaling¹. Defaults to 'lin'.
            cmap (cmap_names, optional): The colormap. Defaults to 'coolwarm'.
            clim (tuple[float, float], optional): Specific color limits (min, max). Defaults to None.
            opacity (float, optional): The opacity of the surface. Defaults to 1.0.
            symmetrize (bool, optional): Wether to force a symmetrical color limit (-A,A). Defaults to True.
        
        (¹): lin: f(x)=x, log: f(x)=log₁₀(|x|), symlog: f(x)=sgn(x)·log₁₀(1+|x·ln(10)|)
        """
        
        grid = pv.StructuredGrid(x,y,z)
        field_flat = field.flatten(order='F')
        
        if scale=='log':
            T = lambda x: np.log10(np.abs(x+1e-12))
        elif scale=='symlog':
            T = lambda x: np.sign(x) * np.log10(1 + np.abs(x*np.log(10)))
        else:
            T = lambda x: x
        
        static_field = T(np.real(field_flat))
        
        if _fieldname is None:
            name = 'anim'+str(self._ctr)
        else:
            name = _fieldname
        self._ctr += 1
        
        has_nan = np.any(np.isnan(static_field))
        if has_nan:
            nan_opacity = 0.0
        else:
            nan_opacity = 1.0
            
        grid[name] = static_field
        
        grid_no_nan = grid.threshold(scalars=name)
        
        default_cmap = self.set.theme.default_amplitude_cmap
        # Determine color limits
        if clim is None:
            if self._cbar_lim is not None:
                clim = self._cbar_lim
            else:
                fmin = np.nanmin(static_field)
                fmax = np.nanmax(static_field)
                clim = (fmin, fmax)
        
        if symmetrize:
            lim = max(abs(clim[0]), abs(clim[1]))
            clim = (-lim, lim)
            default_cmap = self.set.theme.default_wave_cmap
        
        if cmap is None:
            cmap = default_cmap
        else:
            cmap = self.set.theme.parse_cmap_name(cmap)
            
        kwargs = setdefault(kwargs, cmap=cmap, clim=clim, opacity=opacity, pickable=False, multi_colors=True, nan_opacity=nan_opacity)
        actor = self._wrap_plot(grid_no_nan, scalars=name, scalar_bar_args=self._cbar_args, **kwargs)
        
        if self._animate_next:
            def on_update(obj: _AnimObject, phi: complex):
                field_anim = obj.T(np.real(obj.field * phi))
                obj.grid[name] = field_anim
                obj.fgrid[name] = obj.grid.threshold(scalars=name)[name]
                #obj.fgrid replace with thresholded scalar data.
            self._objs.append(_AnimObject(field_flat, T, grid, grid_no_nan, actor, on_update))
            self._animate_next = False
        self._reset_cbar()
        return grid_no_nan



# EMerge is an open source Python based FEM EM simulation module.
# Copyright (C) 2025  Robert Fennis.

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, see
# <https://www.gnu.org/licenses/>.

import re
from typing import Any
from ..pcb import PCB, RouteException
from ...selection import Selection
from ...geometry import GeoObject
from emsutil import Material, PEC
from ...cs import GCS, CoordinateSystem
from loguru import logger
from math import hypot
from ...simmodel import Simulation
try: 
    import ezdxf
    from ezdxf.recover import readfile as recover_readfile
    from ezdxf.path import make_path, from_hatch
except ImportError as e:
    logger.error('Cannot find the required ezdxf library. Install using: pip install ezdxf')
    raise e
import numpy as np

INSUNITS_TO_NAME = {
    0: "unitless",
    1: "inch",
    2: "foot",
    3: "mile",
    4: "millimeter",
    5: "centimeter",
    6: "meter",
    7: "kilometer",
    8: "microinch",
    9: "mil",          # thousandth of an inch
    10: "yard",
    11: "angstrom",
    12: "nanometer",
    13: "micrometer",
    14: "decimeter",
    15: "decameter",
    16: "hectometer",
    17: "gigameter",
    18: "astronomical unit",
    19: "light year",
    20: "parsec",
}

# scale factor: drawing units -> millimeters
INSUNITS_TO_MM = {
    0: 1.0,           # unitless; treat as mm by convention
    1: 25.4,
    2: 304.8,
    3: 1609344.0,
    4: 1.0,
    5: 10.0,
    6: 1000.0,
    7: 1_000_000.0,
    8: 25.4e-6,
    9: 0.0254,
    10: 914.4,
    11: 1e-7,
    12: 1e-6,
    13: 1e-3,
    14: 100.0,
    15: 10_000.0,
    16: 100_000.0,
    17: 1e12,
    18: 1.495978707e14,     # AU in mm
    19: 9.460730472e18,     # ly in mm
    20: 3.085677581e19,     # pc in mm
}

def cluster_values(values, tol):
    """Return [(center, count), ...] clustering sorted values by tolerance."""
    if not values:
        return []
    values = sorted(values)
    clusters = [[values[0]]]
    for v in values[1:]:
        if abs(v - clusters[-1][-1]) <= tol:
            clusters[-1].append(v)
        else:
            clusters.append([v])
    out = []
    for c in clusters:
        out.append((sum(c)/len(c), len(c)))
    return out

STACKUP_TOKENS = [
    r"fr-?4", r"rogers ?\d+", r"core", r"prepreg", r"dielectric",
    r"cu(?:\W|$)|copper", r"\b1\.?6\s*mm\b", r"\b0\.?8\s*mm\b",
    r"\b[12]\s*oz\b", r"\b35\s*µ?m\b", r"\b70\s*µ?m\b",
    r"stack[- ]?up", r"layer\s*\d+", r"pcb", r"thickness",
]

def inspect_pcb_from_dxf(
    filename: str,
    flatten_tol: float = 0.25,    # curve flattening tolerance (drawing units)
    z_tol: float | None = None,   # cluster tolerance for Z (drawing units); default auto
    layer_filter: str | None = None,
):
    """
    Returns a dict with:
      {
        'units': {'code': int, 'name': str, 'to_mm': float},
        'z_levels': [{'z': float, 'count': int}],           # in drawing units
        'thickness_units': 'same as units',
        'thickness': float,                                  # in drawing units
        'thickness_mm': float,                               # convenience
        'notes': {'materials': [...], 'raw_text_hits': [...]}
      }
    """
    doc, auditor = recover_readfile(filename)
    msp = doc.modelspace()

    # Units
    code = int(doc.header.get("$INSUNITS", 0))
    unit_name = INSUNITS_TO_NAME.get(code, f"unknown({code})")
    to_mm = INSUNITS_TO_MM.get(code, 1.0)

    # Gather geometry Zs (WCS), also consider entity.dxf.elevation as fallback
    z_values = []

    def on_layer(e) -> bool:
        return layer_filter is None or e.dxf.layer == layer_filter

    for e in msp:
        if not on_layer(e):
            continue
        dxtype = e.dxftype()
        paths = []
        try:
            if dxtype in ("HATCH", "MPOLYGON"):
                paths = list(from_hatch(e))
            else:
                paths = [make_path(e)]
        except TypeError:
            # skip unsupported entity types
            pass
        except Exception:
            pass

        for p in paths:
            subs = p.sub_paths() if getattr(p, "has_sub_paths", False) and p.has_sub_paths else [p]
            for sp in subs:
                try:
                    verts = list(sp.flattening(distance=flatten_tol))
                except Exception:
                    continue
                for v in verts:
                    # v may be Vec2 or Vec3; handle both
                    z = getattr(v, "z", None)
                    if z is None:
                        # fallback to entity elevation if available
                        z = float(getattr(e.dxf, "elevation", 0.0))
                    z_values.append(float(z))

        # Entities with explicit THICKNESS (rare but possible)
        thick = getattr(e.dxf, "thickness", None)
        if thick not in (None, 0):
            # If an entity is extruded, its "top" would be at elevation+thickness along normal.
            # We can't reliably map OCS normal here; just record the magnitude as a hint.
            pass

    # Cluster Zs
    auto_tol = (1e-6 if to_mm >= 1.0 else 1e-6 / to_mm)  # ~1 nm in mm space → tiny in most units
    ztol = z_tol if z_tol is not None else auto_tol
    z_clusters = cluster_values(z_values, tol=ztol)
    z_levels = [{"z": z, "count": n} for z, n in z_clusters]

    # Thickness from geometry (only meaningful if multiple distinct Zs)
    if z_values:
        zmin, zmax = min(z_values), max(z_values)
        thickness = zmax - zmin
    else:
        zmin = zmax = thickness = 0.0

    # Parse material/thickness hints from TEXT/MTEXT
    material_hits = set()
    raw_text_hits = []
    token_re = re.compile("|".join(STACKUP_TOKENS), re.IGNORECASE)

    for e in msp.query("TEXT MTEXT"):
        try:
            text = e.dxf.text if e.dxftype() == "TEXT" else e.text
        except Exception:
            continue
        if not text:
            continue
        if token_re.search(text):
            raw_text_hits.append(text.strip())
            # simple keyword extraction
            for kw in ["FR4", "Rogers", "core", "prepreg", "copper", "stackup", "thickness", "oz", "µm", "um", "mm"]:
                if re.search(kw, text, re.IGNORECASE):
                    material_hits.add(kw.upper())

    return {
        "units": {"code": code, "name": unit_name, "to_mm": to_mm},
        "z_levels": z_levels,
        "thickness_units": unit_name,
        "thickness": thickness,
        "thickness_mm": thickness * to_mm,
        "notes": {
            "materials": sorted(material_hits),
            "raw_text_hits": raw_text_hits[:50],  # cap to keep output sane
            "comment": (
                "Single Z level detected; geometry alone cannot determine PCB thickness."
                if len(z_levels) <= 1 else
                "Thickness estimated from min/max Z across all geometry."
            )
        },
    }

def path_items_with_semantics(e, p, distance=0.25):
    """Yield dicts describing each sub-path with geometry + semantics we can infer."""
    subs = p.sub_paths() if getattr(p, "has_sub_paths", False) and p.has_sub_paths else [p]
    for sp in subs:
        verts = list(sp.flattening(distance=distance))
        if len(verts) < 2:
            continue

        # geometry in XY (keep Z if you like)
        pts = [(v.x, v.y) for v in verts]
        is_closed = pts[0] == pts[-1]

        item = {
            "layer": e.dxf.layer,
            "entity": e.dxftype(),
            "handle": e.dxf.handle,
            "is_closed": is_closed,
            "points": pts if is_closed else None,       # polygon ring for closed shapes
            "start": None,
            "end": None,
            "length": None,                             # centerline length for open paths
            "width": None,                              # LWPOLYLINE width if available
        }

        if not is_closed:
            item["start"] = pts[0]
            item["end"]   = pts[-1]
            # polyline centerline length
            item["length"] = sum(
                hypot(x2 - x1, y2 - y1) for (x1, y1), (x2, y2) in zip(pts, pts[1:])
            )

        # Preserve potential trace width info from LWPOLYLINE
        if e.dxftype() == "LWPOLYLINE":
            cw = getattr(e.dxf, "const_width", None)
            if cw not in (None, 0):
                item["width"] = float(cw)
            else:
                # per-vertex widths (rare but possible)
                try:
                    item["widths"] = [(v[0], v[1]) for v in zip(e.get_start_widths(), e.get_end_widths())]
                except Exception:
                    pass

        yield item
        
def extract_polygons_with_meta(
    filename: str,
    layer: str | None = None,
    distance: float = 0.25,
    skip_open: bool = True,
) -> dict[str, Any]:
    """
    Returns a list of dicts:
      {
        'ring': [(x, y), ...],         # closed ring
        'layer': 'LayerName',
        'z': 12.34,                    # representative height (median Z of ring)
        'entity': 'LWPOLYLINE' | ...,
        'handle': 'ABCD'
      }
    """
    doc, auditor = recover_readfile(filename)
    msp = doc.modelspace()

    def on_layer(e) -> bool:
        return layer is None or e.dxf.layer == layer

    items = []

    for e in msp:
        if not on_layer(e):
            continue

        dxtype = e.dxftype()
        paths = []
        try:
            if dxtype in ("HATCH", "MPOLYGON"):
                paths = list(from_hatch(e))
            else:
                paths = [make_path(e)]
        except TypeError:
            continue
        except Exception:
            continue

        for p in paths:
            subs = p.sub_paths() if p.has_sub_paths else [p]
            for sp in subs:
                
                verts = list(sp.flattening(distance=distance))  # Vec3 in WCS
                if len(verts) < 2:
                    continue

                # ensure not closed
                if (verts[0].x, verts[0].y) == (verts[-1].x, verts[-1].y):
                    verts = verts[:-1]

                # ring in XY (your original target), but also compute a Z for reference
                ring_xy = [(v.x, v.y) for v in verts]
                zs = sorted(v.z for v in verts)
                z = zs[len(zs)//2] if zs else float(getattr(e.dxf, "elevation", 0.0))  # median Z

                items.append({
                    "ring": ring_xy,
                    "layer": e.dxf.layer,
                    "z": float(z),
                    "entity": dxtype,
                    "handle": e.dxf.handle,
                })

    return items

def import_dxf(filename: str, 
               material: Material, 
               thickness: float | None = None,
               unit: float | None = None,
               cs: CoordinateSystem | None = GCS,
               trace_material: Material = PEC) -> PCB:
    
    polies = extract_polygons_with_meta(filename)
    prop = inspect_pcb_from_dxf(filename)
    
    if prop['units']['name'] == 'unitless':
        if unit is None:
            raise RouteException(f'Cannot generate PCB because the unit is not found in the DXF file or provided in the import_dxf function.')
        pcb_unit = unit
    else:
        pcb_unit = 0.001 * prop['units']['to_mm']
    
    if prop['thickness'] == 0.0:
        if thickness is None:
            raise RouteException(f'Cannot generate PCB because no thickness is found int he DXF file and none is provided in the import_dxf function.')
        pcb_thickness = thickness
    else:
        pcb_thickness = 0.001 * prop['thickness_mm'] / pcb_unit
    
    if cs is None:
        cs = GCS
    
    zs = sorted(list(set([pol['z'] for pol in polies])))
    pcb = PCB(pcb_thickness, pcb_unit, cs, material=material, trace_material=trace_material)
    
    for poly in polies:
        xs, ys = zip(*poly['ring'])
        z = poly['z']
        zs.append(z)
        xs = [x for x in xs]
        ys = [y for y in ys]
        
        pcb.add_poly(xs, ys, z=z, name=poly['handle'])
    return pcb

def _extract_polygons(nodes: np.ndarray, tris: np.ndarray, tri_ids: np.ndarray) -> list[list[tuple[float, float]]]:
    from collections import defaultdict
    edge_counter = defaultdict(int)
    
    nT = tri_ids.shape[1]
    for it in range(nT):
        i1, i2, i3 = [int(x) for x in tri_ids[:,it]]
        edge_counter[(min(i1,i2), max(i1,i2))] += 1
        edge_counter[(min(i1,i3), max(i1,i3))] += 1
        edge_counter[(min(i2,i3), max(i2,i3))] += 1

    edges: list[tuple[int, int]] = [edge for edge, counter in edge_counter.items() if counter == 1]
    
    # stitch edges
    
    node_sequence = defaultdict(list)
    
    for i1, i2 in edges:
        node_sequence[i1].append(i2)
        node_sequence[i2].append(i1)
    
    # Extract sequence loops
    # Node sequence maps node -> [node1, node2] but its unknon if node1 or node2 comes first:
    
    
    islands = []
    
    while True:
        first_node = node_sequence.keys()
        if not first_node:
            break
        first_node = list(first_node)[0]
        second_node = node_sequence[first_node][0]
        
        loop = [first_node, second_node]
        
        while True:
            next_nodes = node_sequence[second_node]
            next_node = next_nodes[0] if next_nodes[0] != first_node else next_nodes[1]
            if next_node == loop[0]:
                break
            loop.append(next_node)
            first_node = second_node
            second_node = next_node
        
        islands.append(loop)
        for node in loop:
            del node_sequence[node]
    
    
    # Create Polygons
    polygons = []
    
    for loop in islands:
        poly = []
        for node in loop:
            x = nodes[0, node]
            y = nodes[1, node]
            poly.append((x,y))
        
        # remove last point if its within 1e-8 of the first
        if hypot(poly[0][0]-poly[-1][0], poly[0][1]-poly[-1][1]) < 1e-8:
            poly = poly[:-1]    
        
        # remove all collinear points
        simplified_poly = []
        nP = len(poly)
        for i in range(nP):
            p_prev = poly[i-1]
            p_curr = poly[i]
            p_next = poly[(i+1)%nP]
            
            # vector from prev to curr
            v1 = (p_curr[0]-p_prev[0], p_curr[1]-p_prev[1])
            # vector from curr to next
            v2 = (p_next[0]-p_curr[0], p_next[1]-p_curr[1])
            
            cross = v1[0]*v2[1] - v1[1]*v2[0]
            if abs(cross) > 1e-10:
                simplified_poly.append(p_curr)
        
        simplified_poly.append(simplified_poly[0])  # close the polygon
        polygons.append(simplified_poly)
    
    return polygons

def _export_dxf_single(simulation: Simulation, filename: str, z_height: float, selection: Selection | GeoObject | list[Selection | GeoObject] | None = None) -> None:
    
    if selection is None:
        selection = []
    elif isinstance(selection, (Selection, GeoObject)):
        selection = [selection]
    
    all_tags = set()
    
    mesh = simulation.mesh
    
    for key in mesh.ftag_to_tri.keys():
        all_tags.add(key)
    
    
    if len(selection) > 0:
        selected_tags = set()
        for sel in selection:
            selected_tags.update(sel.tags)
        # intersect with all_tags
        selected_tags = selected_tags.intersection(all_tags)
    else:
        selected_tags = all_tags
    
    nodes = mesh.nodes
    tris = mesh.tris
    
    poly_triset = []
    for geo in simulation.all_geos():
        if geo.dim != 2:
            continue
        
        tags = [tag for tag in geo.tags if tag in selected_tags]
        if not tags:
            continue
        tri_ids = mesh.get_triangles(tags)
        
        polynodes = nodes[:,np.unique(tris[:,tri_ids])]
        
        if not np.all(np.abs(polynodes[2,:]-z_height) < 1e-8):
            continue

        poly_triset.append(tris[:,tri_ids])
        
    
    # entire tri_set is on the correct z-height
    polies: list[list] = [_extract_polygons(nodes, tris, triset) for triset in poly_triset]
    
    # Reduce polies to a single list of polygons
    final_polies = []
    for polyset in polies:
        for poly in polyset:
            
            final_polies.append(poly)
    
    if not final_polies:
        logger.warning(f'No 2D geometry found at z={z_height} to export to DXF.')
        return
    doc = ezdxf.new(dxfversion='R2010')
    msp = doc.modelspace()
    for poly in final_polies:
        # add as solid region
        
        msp.add_lwpolyline(poly, close=True, dxfattribs={'layer': 'F.Cu'})
    # add .dxf if it is not thetere
    if not filename.lower().endswith('.dxf'):
        filename += '.dxf'
        
    doc.saveas(filename)
    return

def pcb_to_dxf(pcb: PCB, filename: str) -> None:
    """Exports all PCB traces to a DXF file.

    Args:
        PCB (PCB): Your PCB object.
        filename (str): The filename to save the DXF as.
        
    Returns:
        None
    """
    
    from collections import defaultdict
    
    unit = pcb.unit
    layers: dict[float, list] = defaultdict(list)
    for pcbpoly in pcb._poly_out:
        xs = list(np.array(pcbpoly.xs)*unit)
        ys = list(np.array(pcbpoly.ys)*unit)
        poly = [(x,y) for x,y in zip(xs, ys)]
        # make sure the last point is the same as the first
        if poly[0] != poly[-1]:
            poly.append(poly[0])
        layers[pcbpoly.z].append(poly)
    
    for z in layers:
        doc = ezdxf.new(dxfversion='R2010')
        msp = doc.modelspace()
        
        for poly in layers[z]:
            msp.add_lwpolyline(poly, close=True, dxfattribs={'layer': f'Z{z:.3f}'})
        # add .dxf if it is not thetere
        fname_parts = filename.rsplit('.', maxsplit=1)
        if len(fname_parts) == 2:
            fname = f'{fname_parts[0]}_z{z:.3f}.{fname_parts[1]}'
        else:
            fname = f'{filename}_z{z:.3f}.dxf'
            
        doc.saveas(fname)

def export_dxf(simulation: Simulation, filename: str, z_height: float | list[float], selection: list[Selection | GeoObject] = None) -> None:
    """Exports all 2D Geometries on a single or multiple  z-heights to a DXF file.

    Args:
        simulation (Simulation): Your simulation object.
        filename (str): The filename to save the DXF as.
        z_height (float): The z-height to export the 2D geometries at.
        
    Returns:
        None
    """
    
    if not simulation.mesh.defined:
        raise Exception('Cannot export DXF because the simulation mesh is not defined yet. Please run generate_mesh() first.')

    if isinstance(z_height, list):
        for zh in z_height:
            fname_parts = filename.rsplit('.', maxsplit=1)
            if len(fname_parts) == 2:
                fname = f'{fname_parts[0]}_z{zh:.3f}.{fname_parts[1]}'
            else:
                fname = f'{filename}_z{zh:.3f}.dxf'
            _export_dxf_single(simulation, fname, zh, selection)
    else:
        _export_dxf_single(simulation, filename, z_height, selection)
    return


import warnings
import mumps._dmumps
import mumps._zmumps

__all__ = [
    'DMumpsContext',
    'ZMumpsContext',
    'spsolve',
    ]

########################################################################
# Classes
########################################################################

# The main class which is shared between the various datatype variants.
class _MumpsBaseContext(object):
    """MUMPS Context

    This context acts as a thin wrapper around MUMPS_STRUC_C
    which is accessible in the `id` attribute.

    Usage
    -----

    Basic usage generally involves setting up the context, adding
    the sparse matrix and right hand side in process 0, and using
    `run` to execute the various MUMPS phases.

        ctx = MumpsContext()
        if rank == 0:
            ctx.set_centralized_sparse(A)
            x = b.copy() # MUMPS modifies rhs in place, so make copy
            ctx.set_rhs(x)
        ctx.run(6) # Symbolic + Numeric + Solve
        ctx.destroy() # Free internal data structures

        assert abs(A.dot(x) - b).max() < 1e-10
    """

    def __init__(self, par=1, sym=0, comm=None):
        """Create a MUMPS solver context.

        Parameters
        ----------
        par : int
            1 if rank 0 participates
            0 if rank 0 does not participate
        sym : int
            0 if unsymmetric
        comm : MPI Communicator or None
            If None, use MPI_COMM_WORLD
        """
        if comm is None:
            from mpi4py import MPI
            comm = MPI.COMM_WORLD
        self.comm = comm

        self.id = self._MUMPS_STRUC_C()
        self.id.par = par
        self.id.sym = sym
        self.id.comm_fortran = comm.py2f()
        self.run(job = -1) # JOB_INIT
        self.myid = comm.rank
        self._refs = {} # References to matrices

    def __enter__(self):
        return self

    def __exit__(self, *exc_info):
        self.destroy()

    def set_shape(self, n):
        """Set the matrix shape."""
        self.id.n = n

    def set_centralized_sparse(self, A):
        """Set assembled matrix on processor 0.

        Parameters
        ----------
        A : `scipy.sparse.coo_matrix`
            Sparse matrices of other formats will be converted to
            COOrdinate form.
        """
        if self.myid != 0:
            return

        A = A.tocoo()
        n = A.shape[0]
        assert A.shape == (n, n), "Expected a square matrix."
        self.set_shape(n)
        self.set_centralized_assembled(A.row+1, A.col+1, A.data)


    ####################################################################
    # Centralized (the rank 0 process supplies the entire matrix)
    ####################################################################

    def set_centralized_assembled(self, irn, jcn, a):
        """Set assembled matrix on processor 0.

        The row and column indices (irn & jcn) should be one based.
        """
        self.set_centralized_assembled_rows_cols(irn, jcn)
        self.set_centralized_assembled_values(a)

    def set_centralized_assembled_rows_cols(self, irn, jcn):
        """Set assembled matrix indices on processor 0.

        The row and column indices (irn & jcn) should be one based.
        """
        if self.myid != 0:
            return
        assert irn.size == jcn.size
        self._refs.update(irn=irn, jcn=jcn)
        self.id.nz = irn.size
        self.id.irn = self.cast_array(irn)
        self.id.jcn = self.cast_array(jcn)

    def set_centralized_assembled_values(self, a):
        """Set assembled matrix values on processor 0."""
        if self.myid != 0:
            return
        assert a.size == self.id.nz
        self._refs.update(a=a)
        self.id.a = self.cast_array(a)


    ####################################################################
    # Distributed (each process enters some portion of the matrix)
    ####################################################################

    def set_distributed_assembled(self, irn_loc, jcn_loc, a_loc):
        """Set the distributed assembled matrix.

        Distributed assembled matrices require setting icntl(18) != 0.
        """
        self.set_distributed_assembled_rows_cols(irn_loc, jcn_loc)
        self.set_distributed_assembled_values(a_loc)

    def set_distributed_assembled_rows_cols(self, irn_loc, jcn_loc):
        """Set the distributed assembled matrix row & column numbers.

        Distributed assembled matrices require setting icntl(18) != 0.
        """
        assert irn_loc.size == jcn_loc.size

        self._refs.update(irn_loc=irn_loc, jcn_loc=jcn_loc)
        self.id.nz_loc = irn_loc.size
        self.id.irn_loc = self.cast_array(irn_loc)
        self.id.jcn_loc = self.cast_array(jcn_loc)

    def set_distributed_assembled_values(self, a_loc):
        """Set the distributed assembled matrix values.

        Distributed assembled matrices require setting icntl(18) != 0.
        """
        assert a_loc.size == self._refs['irn_loc'].size
        self._refs.update(a_loc=a_loc)
        self.id.a_loc = self.cast_array(a_loc)


    ####################################################################
    # Right hand side entry
    ####################################################################

    def set_rhs(self, rhs):
        """Set the right hand side. This matrix will be modified in place."""
        assert rhs.size == self.id.n
        self._refs.update(rhs=rhs)
        self.id.rhs = self.cast_array(rhs)

    def set_icntl(self, idx, val):
        """Set the icntl value.

        The index should be provided as a 1-based number.
        """
        self.id.icntl[idx-1] = val

    def set_job(self, job):
        """Set the job."""
        self.id.job = job

    def set_silent(self):
        """Silence most messages."""
        self.set_icntl(1, -1) # output stream for error msgs
        self.set_icntl(2, -1) # otuput stream for diagnostic msgs
        self.set_icntl(3, -1) # output stream for global info
        self.set_icntl(4, 0)  # level of printing for errors

    @property
    def destroyed(self):
        return self.id is None

    def destroy(self):
        """Delete the MUMPS context and release all array references."""
        if self.id is not None and self._mumps_c is not None:
            self.id.job = -2 # JOB_END
            self._mumps_c(self.id)
        self.id = None
        self._refs = None

    def __del__(self):
        if not self.destroyed:
            warnings.warn("undestroyed %s" % self.__class__.__name__,
                          RuntimeWarning)
        self.destroy()

    def mumps(self):
        """Call MUMPS, checking for errors in the return code.

        The desired job should have already been set using `ctx.set_job(...)`.
        As a convenience, you may wish to call `ctx.run(job=...)` which sets
        the job and calls MUMPS.
        """
        self._mumps_c(self.id)
        if self.id.infog[0] < 0:
            raise RuntimeError("MUMPS error: %d" % self.id.infog[0])

    def run(self, job):
        """Set the job and run MUMPS.

        Valid Jobs
        ----------
        1 : Analysis
        2 : Factorization
        3 : Solve
        4 : Analysis + Factorization
        5 : Factorization + Solve
        6 : Analysis + Factorization + Solve
        """
        self.set_job(job)
        self.mumps()

class DMumpsContext(_MumpsBaseContext):

    cast_array = staticmethod(_dmumps.cast_array)
    _mumps_c = staticmethod(_dmumps.dmumps_c)
    _MUMPS_STRUC_C = staticmethod(_dmumps.DMUMPS_STRUC_C)

class ZMumpsContext(_MumpsBaseContext):

    cast_array = staticmethod(_zmumps.cast_array)
    _mumps_c = staticmethod(_zmumps.zmumps_c)
    _MUMPS_STRUC_C = staticmethod(_zmumps.ZMUMPS_STRUC_C)

########################################################################
# Functions
########################################################################

def spsolve(A, b, comm=None):
    """Sparse solve A\b."""

    # assert A.dtype == 'd' and b.dtype == 'd', "Only double precision supported."
    if A.dtype == 'd' and b.dtype == 'd':
        with DMumpsContext(par=1, sym=0, comm=comm) as ctx:
            if ctx.myid == 0:
                # Set the sparse matrix -- only necessary on
                ctx.set_centralized_sparse(A.tocoo())
                x = b.copy()
                ctx.set_rhs(x)

            # Silence most messages
            ctx.set_silent()

            # Analysis + Factorization + Solve
            ctx.run(job=6)

            if ctx.myid == 0:
                return x
    elif A.dtype == 'z' or b.dtype == 'z':
        with ZMumpsContext(par=1, sym=0, comm=comm) as ctx:
            if ctx.myid == 0:
                # Set the sparse matrix -- only necessary on
                ctx.set_centralized_sparse(A.tocoo())
                x = b.copy()
                ctx.set_rhs(x)

            # Silence most messages
            ctx.set_silent()

            # Analysis + Factorization + Solve
            ctx.run(job=6)

            if ctx.myid == 0:
                return x


# EMerge is an open source Python based FEM EM simulation module.
# Copyright (C) 2025  Robert Fennis.

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, see
# <https://www.gnu.org/licenses/>.

import numpy as np
from emsutil import Material
from ...const import Z0 as n0

PI = np.pi
TAU = 2*PI

def microstrip_z0(W: float, th: float, er: float):
    u = W/th
    fu = 6 + (TAU - 6)*np.exp(-((30.666/u)**(0.7528)))
    Z0 = n0/(TAU*np.sqrt(er))* np.log(fu/u + np.sqrt(1+(2/u)**2))
    return Z0


class PCBCalculator:

    def __init__(self, thickness: float, layers: np.ndarray, material: Material, unit: float):
        self.th = thickness
        self.layers = layers
        self.mat = material
        self.unit = unit

    def z0(self, Z0: float, layer: int = -1, ground_layer: int = 0, f0: float = 1e9):
        th = abs(self.layers[layer] - self.layers[ground_layer])*self.unit
        ws = np.geomspace(1e-6,1e-1,101)
        Z0ms = microstrip_z0(ws, th, self.mat.er.scalar(f0))
        return np.interp(Z0, Z0ms[::-1], ws[::-1])/self.unit
